(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2007,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* Type checking of more appropriately, consistency checks for
   declared relations.

   Since we generate source code, typechecking problems will be
   reported by the Caml compiler anyway. *)

open Parsetree;;
open Relation;;

type error =
   | Type_not_private of string
   | Unknown_constructor of string
   | Absorbent_and_neutral of string
   | Two_neutral of string
   | Two_absorbent of string
   | Two_in_invol_idem_nil of string
   | Circular_distributivity
   | Type_error of string * relation
   | Combination_not_supported;;

exception Error of Location.t * error;;

(* Simplify a list of relations. *)

let simplify_rels rels = rels;; (* FIXME ? *)

(* Builds a table constructor name / constructor definition. *)

type cinfo = {
  ci_type_name : string;
  ci_arg_types : core_type list;
  ci_relations : relations;
  ci_location : Location.t;
};;

let cname_tbl = Hashtbl.create 97;;

let fill_cname_tbl =
  let add_ntd (tname, td) =
    match td.ptype_kind with
      | Ptype_variant (cds, _) ->
	  let add_cd (cname, cts, rels, loc) =
	    Hashtbl.add cname_tbl cname
	      { ci_type_name = tname;
		ci_arg_types = cts;
		ci_relations = rels;
		ci_location = loc; }
	  in List.iter add_cd cds
      | _ -> ()
  in List.iter add_ntd;;

let find_cname = Hashtbl.find cname_tbl;;

(* Builds a table type name / constructors. *)

let tname_tbl = Hashtbl.create 19;;

let fill_tname_tbl =
  let add_ntd (tname, td) = Hashtbl.add tname_tbl tname td in
  List.iter add_ntd;;

let find_tname = Hashtbl.find tname_tbl;;

(* Check that arguments of relations refer to defined constructors. *)

let check_constr_names =
  let check_name s =
    if not (Hashtbl.mem cname_tbl s)
    then raise (Error ((find_cname s).ci_location, Unknown_constructor s)) in
  let check_rel = function
    | Absorbent (_, s)
    | Absorbing (_, s)
    | Neutral (_, s)
    | Nilpotent s
    | Inverse (_, s, None)
    | Distributive_Unary (s, None)
    | Distributive_Inverse (_, s)
    | Distributive (_, s)
    | Inverse_Neutral (s, None) -> check_name s
    | Inverse (_, s1, Some s2)
    | Distributive_Unary (s1, Some s2)
    | Inverse_Neutral (s1, Some s2) -> check_name s1; check_name s2
    | _ -> ()
  in List.iter check_rel;;

(* Check that no constructor is:
- both absorbent and neutral with the same argument,
- absorbent and absorbent with same side and distinct arguments,
- neutral and neutral with same side and distinct arguments *)

(*let check_abs_neu =
  let is_absorbent_or_neutral = function
    | Absorbent _ | Neutral _ -> true
    | _ -> false in
  let rec check rels =
    match List.filter is_absorbent_or_neutral rels with
      | [] | [_] -> ()
      | Absorbent (_, s) :: Absorbent (_, s') :: _ ->
	  raise (Error (Location.none, Two_absorbent cname))
      | Neutral (_, s) :: Neutral (_, s') :: _ ->
	  raise (Error (Location.none, Two_neutral cname))
      | Absorbent (_, s) :: Neutral (_, s') :: _
      | Neutral (_, s) :: Absorbent (_, s') :: _ when s=s' ->
	  raise (Error (Location.none, Absorbent_and_neutral cname))
      | _ :: rest -> check_lcr cname rest
  in check;;

let check_abs_neu  = check_lcr cname x.ci_relations
*)

(* Proceed to all checks. *)

let sigtype ntds =
  fill_cname_tbl ntds;
  fill_tname_tbl ntds;
  let check _cname ci =
    check_constr_names ci.ci_relations
  in Hashtbl.iter check cname_tbl;;

(* Error messages. *)

open Printf;;

let bprint_error b loc = function
  | Unknown_constructor s ->
    bprintf b "Unknown constructor %s." s
  | Absorbent_and_neutral s ->
    bprintf b
      "Constructor %s has an argument that is both absorbent and neutral." s
  | Two_neutral s ->
    bprintf b "Constructor %s has two neutral arguments." s
  | Two_absorbent s ->
    bprintf b "Constructor %s has two absorbent arguments." s
  | Two_in_invol_idem_nil s ->
    bprintf b
      "Constructor %s has more than one property among Involutive, \
      Idempotent and Nilpotent." s
  | Circular_distributivity ->
    bprintf b "Distributivity properties are circular."
  | Type_error (cname, rel) ->
    bprintf b
      "%s is not compatible with the type of %s."
        (name_of_relation rel) cname
  | Type_not_private s ->
    bprintf b "Type %s is not declared as private." s
  | Combination_not_supported ->
    bprintf b "Combination of equational theories not supported.";;

let pr_error = function
  | Error (loc, e) ->
    let b = Buffer.create 128 in
    bprint_error b loc e;
    prerr_endline (Buffer.contents b)
  | e -> raise e;;




(* old code:

(* precondition for the following: relations are not redundant *)
(* FIXME: really used? *)

(* Check 3: no constructor are more than 1 property in
Involutive, Idempotent and Nilpotent *)

let rec check_3_iter cname x =
  let is_involutive_idem_or_nilpotent =
    function Involutive | Idempotent _ | Nilpotent _ -> true | _ -> false in
  match List.filter is_involutive_idem_or_nilpotent x.ci_relations with
  | [] | [ _ ] -> ()
  | _ -> raise (Error (Location.none, Two_in_invol_idem_nil cname));;

(* check 4: well-foundedness of distributivity *)

let check_succs t cname =
  let rec chk dname =
    try
      let succs = Hashtbl.find t dname in
      if List.mem cname succs
      then raise (Error (Location.none, Circular_distributivity));
      List.iter chk succs
    with Not_found -> () in
  chk;;

let check_cyclic_dist =
  let t = Hashtbl.create 7 in
  fun cname (_s, ci) ->
    let l = distributive ci.ci_relations in
    List.iter (check_succs t cname) l;
    Hashtbl.add t cname l;;

(* check 5: compatibility of relations with type declarations

c Involutive => c:t->t
c Extern_Involution => c : 'a * t -> t
                       c (x, c (..., c (x, y))) = c (x, y)
c Idempotent => c:t -> t
c Nilpotent(d) => c:t -> t & d:t

c Commutative => c: t list -> t
c Associative => c: t list -> t
c Neutral(d) => c: t list -> t & d: t
c Absorbent(d) => c: t list -> t & d: t
c Distributive(d) => c: t list -> t & d: t list -> t
*)

let curry_type x =
  match x.ci_arg_types with
  | [] -> const_type x.ci_type_name
  | [ct] -> arrow_type ct (const_type x.ci_type_name)
  | cts -> arrow_type (tuple_type cts) (const_type x.ci_type_name);;

let get_type dname = curry_type (find_cname dname);;

let error cname rel = raise (Error (Location.none, Type_error (cname, rel)));;

let check_type_rel cname ci =
  let typ = const_type ci.ci_type_name in
  let ltyp = list_type typ in
  function rel ->
    match rel with
    | Involutive | Idempotent _ ->
      (match ci.ci_arg_types with
       | [ct] when ctrel ct typ -> ()
       | _ -> error cname rel)
    | Nilpotent dname ->
      (match ci.ci_arg_types with
       | [ct] when ctrel ct typ && ctrel (get_type dname) typ -> ()
       | _ -> error cname rel)
    | Commutative _ | Associative _ ->
      (match ci.ci_arg_types with
       | [ct] when ctrel ct ltyp -> ()
       | _ -> error cname rel)
    | Neutral (_, dname) | Absorbent (_, dname) ->
      (match ci.ci_arg_types with
       | [ct] when ctrel ct ltyp && ctrel (get_type dname) typ -> ()
       | _ -> error cname rel)
    | Distributive (_, dname) ->
      (match ci.ci_arg_types with
       | [ct] when ctrel ct ltyp
         && ctrel (get_type dname) (arrow_type ltyp typ) -> ()
       | _ -> error cname rel)
    | Idempotent_Unary
    | Idempotent_AC
    | Division_by_Absorbent _
    | Inverse_Neutral _
    | Distributive_Inverse (_, _)
    | Distributive_Unary (_, _)
    | Inverse (_, _, _)
    | Absorbing (_, _)
    | Rewrite (_, _)
      -> (* FIXME *) ();;

let check_type cname ci =
  List.iter (check_type_rel cname ci) ci.ci_relations;;
*)
