(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2007,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: pr_otype.ml,v 1.8 2007/04/26 17:26:41 blanqui Exp $ *)

(* Printing Objective Caml type definitions. *)

open Asttypes;;
open Parsetree;;
open Longident;;
open Printf;;
open Lexing;;
open Location;;

let rec fmt_longident b = function
  | Lident s -> bprintf b "%s" s
  | Ldot (t, s) -> bprintf b "%a.%s" fmt_longident t s
  | Lapply (t1, t2) ->
    bprintf b "%a(%a)" fmt_longident t1 fmt_longident t2;;

let constant b = function
  | Const_int i ->
    if i < 0 then bprintf b "(%i)" i else bprintf b "%i" i
  | Const_char c -> bprintf b "%C" c
  | Const_string s -> bprintf b "%S" s
  | Const_float s -> bprintf b "%s" s
  | Const_int32 li -> bprintf b "%li" li
  | Const_int64 lli -> bprintf b "%Li" lli
  | Const_nativeint ni -> bprintf b "%ni" ni;;

let rec core_type b x =
  match x.ptyp_desc with
  | Ptyp_any -> bprintf b "_"
  | Ptyp_var (s) -> bprintf b "'%s" s
  | Ptyp_arrow (l, ct1, ct2) ->
      bprintf b "(%a -> %a)" core_type ct1 core_type ct2
  | Ptyp_tuple [] -> ()
  | Ptyp_tuple (ct1 :: cts) ->
      bprintf b "(%a * %a)"
        core_type ct1
        (fun b -> List.iter (fun ct -> bprintf b " * %a" core_type ct)) cts
  | Ptyp_constr (li, []) ->
      bprintf b "%a" fmt_longident li
  | Ptyp_constr (li, (ct1 :: cts)) ->
      bprintf b "(%a%a) %a"
        core_type ct1
        (fun b -> List.iter (fun ct -> bprintf b ", %a" core_type ct)) cts
        fmt_longident li
(*| Ptyp_variant (_l, _closed, _low)
  | Ptyp_object _l
  | Ptyp_class (_li, _l, _low)
  | Ptyp_alias (_ct, _s)
  | Ptyp_poly (_sl, _ct)*)
  | _ -> failwith "Not yet implemented";;

let x_core_types b = function
  | [] -> ()
  | [tyc] -> core_type b tyc
  | tyc :: tycl ->
    bprintf b "%a%a" core_type tyc
      (fun b -> List.iter (fun tyc -> bprintf b " * %a" core_type tyc))
      tycl;;

let rec pattern b { ppat_desc = p } =
  match p with
  | Ppat_any -> bprintf b "_"
  | Ppat_var s -> bprintf b "%s" s
  | Ppat_alias (p, s) -> bprintf b "(%a as %s)" pattern p s
  | Ppat_constant c -> constant b c
  | Ppat_tuple ([] | [_]) -> assert false
  | Ppat_tuple (p :: pats) ->
    bprintf b "(%a%a)" pattern p
      (fun b -> List.iter (fun p -> bprintf b ", %a" pattern p)) pats
  | Ppat_construct (id, None, _) -> fmt_longident b id
  | Ppat_construct (id, Some p, _) ->
    bprintf b "(%a %a)" fmt_longident id pattern p
  | Ppat_variant (label, None) -> bprintf b "`%s" label
  | Ppat_variant (label, Some p) ->
    bprintf b "`%s %a" label pattern p
  | Ppat_record idpats ->
    bprintf b "{%a}"
      (fun b ->
         List.iter
           (fun (id, pat) ->
              bprintf b "%a = %a; " fmt_longident id pattern pat))
      idpats
  | Ppat_array pats ->
    bprintf b "[|%a|]"
      (fun b -> List.iter (fun pat -> bprintf b "%a; " pattern pat))
      pats
  | Ppat_or (p1, p2) ->
    bprintf b "(%a | %a)" pattern p1 pattern p2
  | Ppat_constraint (p, ct) ->
    bprintf b "(%a : %a)" pattern p core_type ct
  | Ppat_type id -> bprintf b "#%a" fmt_longident id;;

let value_description b x =
  bprintf b " : %a" core_type x.pval_type;
  match x.pval_prim with
  | [] -> ()
  | l ->
    bprintf b " =\n  %a" (fun b -> List.iter (fun s -> bprintf b " %S" s)) l;;

let ptype_param b p = bprintf b "'%s" p;;

let ptype_params b = function
  | [] -> ()
  | [p] -> bprintf b "%a " ptype_param p
  | p :: pl ->
    bprintf b "(%a%a)" ptype_param p
      (fun b -> List.iter (fun p -> bprintf b ", %a" ptype_param p))
      pl;;

let side_spec b = function
  | Right -> bprintf b "right"
  | Left -> bprintf b "left"
  | Both -> bprintf b "both";;

let side b sd =
  match sd with
  | Right
  | Left -> bprintf b " %a" side_spec sd
  | Both -> ();;

let find_text_of_expression e =
  let loc = e.pexp_loc in
  let loc_start = loc.loc_start in
  let fname, _, _ = get_pos_info loc_start in
  let pos_cnum = loc_start.pos_cnum in
  let pos_enum = loc.loc_end.pos_cnum in
  let ic =
    try open_in fname with
    | _ ->
      failwith
        (Printf.sprintf
           "find_text_of_expression: \
            file ``%s'' not found or not readable." fname) in
  let len = pos_enum - pos_cnum in
  let b = String.create len in
  try
    seek_in ic pos_cnum;
    really_input ic b 0 len;
    b with
  | End_of_file ->
    failwith
      (Printf.sprintf
         "find_text_of_expression: \
          file ``%s'' not found or not readable." fname)
;;

let expression b e = bprintf b "%s" (find_text_of_expression e);;

let relation b = function
  | Rewrite (p1, None, p2) -> bprintf b "rule %a -> %a" pattern p1 pattern p2
  | Rewrite (p1, Some e, p2) ->
      bprintf b "rule %a when %a -> %a" pattern p1 expression e pattern p2
  | Commutative sd -> bprintf b "commutative%a" side sd
  | Associative sd -> bprintf b "associative%a" side sd
  | Absorbent (sd, cid) -> bprintf b "absorbent%a (%s)" side sd cid
  | Absorbing (sd, cid) -> bprintf b "absorbing%a (%s)" side sd cid
  | Neutral (sd, cid) -> bprintf b "neutral%a (%s)" side sd cid
  | Idempotent_Unary -> bprintf b "idempotent unary"
  | Idempotent_AC -> bprintf b "idempotent AC"
  | Idempotent sd -> bprintf b "idempotent%a" side sd
  | Nilpotent cid -> bprintf b "nilpotent (%s)" cid
  | Involutive -> bprintf b "involutive"
  | Inverse (sd, cid1, None) ->
    bprintf b "inverse%a (%s)" side sd cid1
  | Inverse (sd, cid1, Some cid2) ->
    bprintf b "inverse%a (%s, %s)" side sd cid1 cid2
  | Distributive (sd, cid) -> bprintf b "distributive%a (%s)" side sd cid
  | Distributive_Inverse (sd, cid) ->
    bprintf b "distributive inverse%a (%s)" side sd cid
  | Distributive_Unary (cid1, None) ->
    bprintf b "unary distributive (%s)" cid1
  | Distributive_Unary (cid1, Some cid2) ->
    bprintf b "unary distributive (%s, %s)" cid1 cid2
  | Inverse_Neutral (cid1, None) ->
    bprintf b "inverse neutral (%s)" cid1
  | Inverse_Neutral (cid1, Some cid2) ->
    bprintf b "inverse neutral (%s, %s)" cid1 cid2
  | Division_by_Absorbent _ -> bprintf b "division by absorbent";;

let relations b = function
  | [] -> ()
  | rls ->
    bprintf b "\nbegin%aend\n"
      (fun b -> List.iter (fun rl -> bprintf b "\n%a\n" relation rl)) rls;;

let constructor b (cid, tycl, rels, _loc) =
  match tycl with
  | [] -> bprintf b "\n| %s%a" cid relations rels
  | _ -> bprintf b "\n| %s of %a%a" cid x_core_types tycl relations rels
;;

let constructors b cstrs =
  List.iter (fun cstr -> bprintf b "%a" constructor cstr) cstrs;;
 
let private_flag b = function
  | Public -> ()
  | Private -> bprintf b " private";;

let mutable_flag b = function
  | Immutable -> ()
  | Mutable -> bprintf b "mutable ";;

let field b (s, mflag, cty, _loc) =
  bprintf b "%a%s : %a" mutable_flag mflag s core_type cty;;

let fields b = function
  | [] -> ()
  | fld :: flds ->
    bprintf b "%a%a" field fld
      (fun b -> List.iter (fun fld -> bprintf b ";\n%a" field fld)) flds;;

let rec type_declaration s b ty_decl =
  bprintf b "%a%s" ptype_params ty_decl.ptype_params s;
  (* Skip variance *)
  (* Skip constraints *)
  ptype_manifest b ty_decl.ptype_manifest;
  ptype_kind b ty_decl.ptype_kind

and ptype_manifest b = function
  | None -> ()
  | Some cty -> bprintf b " = %a" core_type cty

and type_declarations b s_ty_decls =
  List.iter
    (fun (s, ty_decl) ->
       bprintf b "\nand %a" (type_declaration s) ty_decl) s_ty_decls

and ptype_kind b = function
  | Ptype_abstract (relations, pflag) -> ()
  | Ptype_variant (cstrs, pflag) ->
      bprintf b " =%a%a" private_flag pflag constructors cstrs
  | Ptype_record (flds, pflag) ->
      bprintf b " =%a{\n%a\n}" private_flag pflag fields flds

and pstr_type b = function
  | [s, ty_decl] -> type_declaration s b ty_decl
  | (s, ty_decl) :: s_ty_decls ->
    bprintf b "%a%a" (type_declaration s) ty_decl
     type_declarations s_ty_decls
  | _ -> ();;
